from typing import AsyncGenerator, Mapping, Optional
from motor.motor_asyncio import AsyncIOMotorClient


class MongoWrapper:
    """Doesn't cover mongo syntax entirely, but just
       includes needed functions.
    """

    def __init__(self, db: AsyncIOMotorClient) -> None:
        self._db = db

    async def exists(self, table: str, or_: dict) -> bool:
        return await self._db[table].count_documents(
            {"$or": [{key: value} for key, value in or_.items()]}
        ) != 0

    async def delete(self, table: str,
                     and_: dict) -> None:
        await self._db[table].delete_many(and_)

    async def update(self, table: str,
                     and_: dict, values: dict) -> None:
        await self._db[table].update_one(
            and_,
            {"$set": values}
        )

    async def insert(self, table: str,
                     values: dict) -> None:
        await self._db[table].insert_one(values)

    async def get(self, table, and_: dict) -> Optional[Mapping]:
        return await self._db[table].find_one(and_)

    async def iterate(self, table: str, and_: dict = None
                      ) -> AsyncGenerator[Mapping, None]:
        find = self._db[table].find(and_) if and_ else self._db[table].find()

        async for document in find:
            yield document



from databases import Database


class SQLEngine:
    def __init__(self, connection: Database) -> None:
        """For SQL database engines.

        postgresql, mysql, & sqlite supported.

        Parameters
        ----------
        connection : Database
        """

        self._connection = connection


class MongoEngine:
    def __init__(self, host: str = "localhost",
                 port: int = 27017, database: str = "aioaccount"
                 ) -> None:
        """For mongodb engine.

        Parameters
        ----------
        host : str, optional
            by default "localhost"
        port : int, optional
            by default 27017
        """

        self._connection = f"mongodb://{host}:{port}"
        self._database = database



from .base import TestBase
from .. import UserModel, UserIdError


class TestGetUser(TestBase):
    use_sql = True

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_invalid_id(self) -> None:
        with self.assertRaises(UserIdError):
            await self.handler.user("123").get()

    async def test_valid_get(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            name="hhg"
        )

        self.assertIsInstance(
            await user.get(),
            UserModel
        )


class TestGetUserMongo(TestGetUser):
    use_sql = False



import asynctest

from .. import (
    AccountHandler,
    MongoEngine,
    SQLEngine,
    Database,
    SmtpClient
)

from .args import MONGO_SETTINGS, SQL_CONNECTION, SMTP_SETTINGS


class TestBase(asynctest.TestCase):
    use_default_loop = True
    use_sql = False
    use_smtp = False

    handler: AccountHandler

    async def setUp(self) -> None:
        engine = MongoEngine(
            **MONGO_SETTINGS
        ) if not self.use_sql else SQLEngine(
            Database(SQL_CONNECTION)
        )
        self.handler = AccountHandler(
            engine=engine,
            smtp=SmtpClient(
                **SMTP_SETTINGS
            ) if self.use_smtp else None
        )

        await self.handler.start()

    async def tearDown(self) -> None:
        await self.handler.close()



import aiojobs

from typing import AsyncGenerator, Tuple, Union
from databases import Database
from secrets import token_urlsafe
from motor.motor_asyncio import AsyncIOMotorClient
from password_strength import PasswordPolicy as ExtPP
from email_validator import validate_email, EmailNotValidError
from bcrypt import hashpw, gensalt, checkpw
from datetime import timedelta

from ._engines import SQLEngine, MongoEngine
from ._smtp import SmtpClient, SmtpHtml
from ._pass_policy import PasswordPolicy
from ._sql import create_tables, SqlWrapper
from ._mongo import MongoWrapper
from ._errors import (
    AioAccountError,
    AccountDetailsError,
    EmailError,
    DetailsExistError,
    PasswordPolicyError,
    NameLengthInvalidError,
    InvalidLogin,
    UnableToConfirmEmail,
    NameInvalidCharactersError,
    UserIdError,
    PasswordResetInvalid
)
from ._util import generate_id
from ._models import UserModel
from ._user import User
from ._const import _MAX_NAME_LEN

__version__ = "0.0.2"
__url__ = "https://aioaccount.readthedocs.io/en/latest/"
__description__ = "Utility for user account creation."
__author__ = "WardPearce"
__author_email__ = "wardpearce@protonmail.com"
__license__ = " AGPL-3.0"

__all__ = [
    "Database",
    "AccountHandler",
    "SQLEngine",
    "MongoEngine",
    "SmtpClient",
    "PasswordPolicy",
    "AioAccountError",
    "AccountDetailsError",
    "EmailError",
    "DetailsExistError",
    "UserModel",
    "PasswordPolicyError",
    "NameLengthInvalidError",
    "UnableToConfirmEmail",
    "NameInvalidCharactersError",
    "UserIdError",
    "PasswordResetInvalid",
    "InvalidLogin",
    "SmtpHtml"
]


class AccountHandler:
    _smtp: Union[SmtpClient, None]
    _db: Union[Database, AsyncIOMotorClient]
    _jobs: aiojobs.Scheduler
    _policy: ExtPP
    _db_wrapper: Union[SqlWrapper, MongoWrapper]

    def __init__(self, engine: Union[MongoEngine, SQLEngine],
                 password_policy: PasswordPolicy = PasswordPolicy(),
                 smtp: SmtpClient = None,
                 password_reset_expires: timedelta = timedelta(hours=24)
                 ) -> None:
        """Configure how the account handler works.

        Parameters
        ----------
        engine : Union[MongoEngine, SQLEngine]
            Engine for storing accounts.
        password_policy : PasswordPolicy, optional
            Password policies, by default PasswordPolicy()
        smtp : SmtpClient, optional
            Smtp client for email validation, by default None
        password_reset_expires : timedelta, optional
            Amount of time until a password reset expires,
            by default timedelta(hours=24)
        """

        if isinstance(engine, SQLEngine):
            self._db = engine._connection
            self._db_wrapper = SqlWrapper(self._db)

            create_tables(str(engine._connection.url))
        else:
            self._db = AsyncIOMotorClient(
                engine._connection
            )[engine._database]
            self._db_wrapper = MongoWrapper(self._db)

        self._smtp = smtp
        self._policy = password_policy._policy
        self._password_reset_expires = password_reset_expires

    async def start(self) -> None:
        """Opens needed sessions.
        """

        if isinstance(self._db, Database):
            await self._db.connect()

        self._jobs = await aiojobs.create_scheduler()

    async def close(self) -> None:
        """Closes any closed sessions.
        """

        if isinstance(self._db, Database):
            await self._db.disconnect()

        await self._jobs.close()

    async def users(self, email_confirmed: bool = None
                    ) -> AsyncGenerator[Tuple[UserModel, User], None]:
        """Used to list users.

        Parameters
        ----------
        email_confirmed : bool, optional
            by default None

        Yields
        -------
        UserModel
        User
        """

        if email_confirmed is not None:
            and_ = {
                "email_confirmed": None if email_confirmed is False else True
            }
        else:
            and_ = None

        async for result in self._db_wrapper.iterate("user", and_):
            yield (
                UserModel(**result),
                self.user(result["user_id"])
            )

    def user(self, user_id: str) -> User:
        """Used to interact with user.

        Parameters
        ----------
        user_id : str

        Returns
        -------
        User
        """

        return User(self, user_id)

    def _validate_details(self, name: str = None,
                          password: str = None) -> None:
        """Used to validate details.

        Parameters
        ----------
        name : str, optional
        password : str, optional

        Raises
        ------
        NameLengthInvalidError
        NameInvalidCharactersError
        PasswordPolicyError
        """

        if name:
            name_len = len(name)
            if name_len > _MAX_NAME_LEN or name_len < 3:
                raise NameLengthInvalidError(
                    f"Name is over {_MAX_NAME_LEN} or below 3 characters."
                )

            if not name.isalnum():
                raise NameInvalidCharactersError(
                    "Account name can only contain alpha characters."
                )

        if password:
            results = self._policy.test(password)
            if results:
                raise PasswordPolicyError(results)

    async def login(self, password: str, name: str = None,
                    email: str = None, require_email_confirmed: bool = True
                    ) -> Tuple[UserModel, User]:
        """Used to validate user's login.

        Parameters
        ----------
        password : str
        name : str, optional
            by default None
        email : str, optional
            by default None
        require_email_confirmed : bool, optional
            If true email must be confirmed, by default True

        Returns
        -------
        UserModel
            Holds info on user.
        User
            Used to interact with user.

        Raises
        ------
        InvalidLogin
            Raised when user login is invalid.
        AccountDetailsError
        """

        if name:
            search = {
                "name": name
            }
        elif email:
            search = {
                "email": email
            }
        else:
            raise AccountDetailsError("User or email must be provided.")

        result = await self._db_wrapper.get("user", search)
        if not result:
            raise InvalidLogin()

        if (require_email_confirmed and self._smtp and result["email"]
                and not result["email_confirmed"]):
            raise InvalidLogin()

        if not checkpw(password.encode(), result["password"]):
            raise InvalidLogin()

        return UserModel(**result), self.user(result["user_id"])

    def _email_regenerate(self) -> dict:
        return {
            "email_vaildate": token_urlsafe(32),
            "email_confirmed": False
        }

    async def confirm_email(self, email: str, given_code: str) -> User:
        """Used to confirm email from given code.

        Parameters
        ----------
        email : str
            User's email
        given_code : str
            Code the user provided.

        Returns
        -------
        User

        Raises
        ------
        UnableToConfirmEmail
        """

        result = await self._db_wrapper.get("user", {"email": email})
        if not result:
            raise UnableToConfirmEmail()

        # Midigates timing attacks.
        valid = checkpw(
            given_code.encode(), hashpw(
                result["email_vaildate"].encode(),
                gensalt()
            )
        )
        if not valid:
            raise UnableToConfirmEmail()

        await self._db_wrapper.update(
            "user",
            {"user_id": result["user_id"]},
            {
                "email_vaildate": None,
                "email_confirmed": True
            }
        )

        return self.user(result["user_id"])

    async def to_user(self, email: str = None,
                      name: str = None) -> Tuple[UserModel, User]:
        """Used to convert email or user to user object.

        Parameters
        ----------
        email : str, optional
            by default None
        name : str, optional
            by default None

        Returns
        -------
        UserModel
        User

        Raises
        ------
        AccountDetailsError
        """

        values = {}
        if email:
            values["email"] = email
        elif name:
            values["name"] = name
        else:
            raise AccountDetailsError("User or email must be provided.")

        result = await self._db_wrapper.get("user", values)
        if not result:
            raise AccountDetailsError("No user found with those details.")

        return UserModel(**result), self.user(result["user_id"])

    async def create_account(self, password: str,
                             email: str = None, name: str = None
                             ) -> Tuple[UserModel, User]:
        """Used to create a user account.

        Parameters
        ----------
        name : str, optional
            Unique name of user, max length 128
        password : str
            Password of user
        email : str, optional
            by default None

        Returns
        -------
        UserModel
            Holds info on user.
        User
            Used to interact with user.

        Raises
        ------
        DetailsExistError
            Raised when user with given details already exists.
        EmailError
            Raised when email is invalid.
        PasswordPolicyError
            Raised when password doesn't meet password policy
        NameLengthInvalidError
            Raised when name over 128 characters.
        AccountDetailsError
        """

        if not name and not email:
            raise AccountDetailsError("User or email must be provided.")

        self._validate_details(
            name=name,
            password=password
        )

        values = {
            "user_id": generate_id()
        }

        if name:
            values["name"] = name

        user_modal = UserModel(**values)
        if email:
            try:
                valid = validate_email(email)
            except EmailNotValidError as error:
                raise EmailError(str(error))

            values["email"] = valid.email

            if await self._db_wrapper.exists("user", values):
                raise DetailsExistError()

            user_modal.email = values["email"]

            if self._smtp:
                user_modal.email_confirmed = False
                values = {
                    **values,
                    **self._email_regenerate()
                }

                await self._jobs.spawn(
                    self._smtp._send(
                        values["email"],
                        values["email_vaildate"],
                        "confirm"
                    )
                )
        else:
            if await self._db_wrapper.exists("user", values):
                raise DetailsExistError()

        values["password"] = hashpw(  # type: ignore
            password.encode(), gensalt()
        )

        await self._db_wrapper.insert("user", values)

        return user_modal, self.user(values["user_id"])



from .base import TestBase
from .. import EmailError, DetailsExistError
from typing import cast


class TestUpdateEmail(TestBase):
    use_sql = True
    use_smtp = True

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_invalid_email(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            name="lloop"
        )

        with self.assertRaises(EmailError):
            await user.update_email("kkd@@pm.me")

    async def test_details_not_exist(self) -> None:
        model, user = await self.handler.create_account(
            password=self.valid_password,
            email="iio@pm.me"
        )

        with self.assertRaises(DetailsExistError):
            await user.update_email(
                new_email=cast(str, model.email)
            )

    async def test_valid_update_email(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            email="llw3@pm.me"
        )

        await user.update_email(new_email="h3h3@pm.me")

        result = await self.handler._db_wrapper.get(
            "user",
            {"user_id": user.user_id}
        )
        if not result:
            raise Exception("Shouldn't happen")

        self.assertFalse(result["email_confirmed"])
        self.assertIsInstance(result["email_vaildate"], str)


class TestUpdateEmailMongo(TestUpdateEmail):
    use_sql = False
    use_smtp = True



from .base import TestBase
from .. import UserIdError


class TestDeleteAccount(TestBase):
    use_sql = True

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_delete_account(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            name="deleteme"
        )

        await user.delete()

        with self.assertRaises(UserIdError):
            await user.get()


class TestDeleteAccountMongo(TestDeleteAccount):
    use_sql = False



import argparse

cli = argparse.ArgumentParser()

cli.add_argument("--sql_username", type=str, default="")
cli.add_argument("--sql_password", type=str, default="")
cli.add_argument("--sql_server", type=str, default="localhost")
cli.add_argument("--sql_port", type=int, default=3306)
cli.add_argument("--sql_database", type=str, default="")

cli.add_argument("--smtp_hostname", type=str, default="localhost")
cli.add_argument("--smtp_port", type=int, default=25)
cli.add_argument("--smtp_email", type=str, default="")

cli.add_argument("--mongo_server", type=str, default="localhost")
cli.add_argument("--mongo_port", type=int, default=27017)

args = vars(cli.parse_args())

MONGO_SETTINGS = {
    "host": args["mongo_server"],
    "port": args["mongo_port"]
}

SQL_CONNECTION = "mysql://{}:{}@{}:{}/{}?charset=utf8mb4".format(
    args["sql_username"],
    args["sql_password"],
    args["sql_server"],
    args["sql_port"],
    args["sql_database"]
)

SMTP_SETTINGS = {
    "host": args["smtp_hostname"],
    "port": args["smtp_port"],
    "email": args["smtp_email"]
}



from .base import TestBase
from .. import PasswordResetInvalid


class TestPasswordResets(TestBase):
    use_sql = True
    use_smtp = True

    valid_password = "S]Q}67=uLetG{r,_8{"

    async def test_password_confirm_invalid_code(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            name="oewup"
        )

        self.assertIsInstance(await user.reset_password(), str)

        with self.assertRaises(PasswordResetInvalid):
            await user.password_confirm(
                new_password=self.valid_password,
                given_code="ddd"
            )

    async def test_password_no_reset(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            name="hhd"
        )

        with self.assertRaises(PasswordResetInvalid):
            await user.password_confirm(
                new_password=self.valid_password,
                given_code="ddd"
            )

    async def test_invalid_user(self) -> None:
        with self.assertRaises(PasswordResetInvalid):
            await self.handler.user("123").password_confirm(
                new_password=self.valid_password,
                given_code="ddd"
            )

    async def test_valid_code(self) -> None:
        _, user = await self.handler.create_account(
            password=self.valid_password,
            name="xxz"
        )

        code = await user.reset_password()

        await user.password_confirm(
            new_password=self.valid_password,
            given_code=code
        )


class TestPasswordResetsMongo(TestPasswordResets):
    use_sql = False
    use_smtp = False



from .test_create_account import TestCreateAccount, TestCreateAccountSqlSmtp
from .test_to_user import TestToUser, TestToUserSqlSmtp
from .test_confirm_email import TestEmailConfirm, TestEmailConfirmMongo
from .test_login import TestLogin, TestLoginSqlSmtp
from .test_update_password import TestUpdatePassword, TestUpdatePasswordSqlSmtp
from .test_update_email import TestUpdateEmail, TestUpdateEmailMongo
from .test_get_user import TestGetUser, TestGetUserMongo
from .test_password_resets import TestPasswordResets, TestPasswordResetsMongo
from .test_users import TestUsers, TestUsersMongo
from .test_delete_account import TestDeleteAccount, TestDeleteAccountMongo

__all__ = [
    "TestCreateAccount",
    "TestCreateAccountSqlSmtp",
    "TestToUser",
    "TestToUserSqlSmtp",
    "TestEmailConfirm",
    "TestEmailConfirmMongo",
    "TestLogin",
    "TestLoginSqlSmtp",
    "TestUpdatePassword",
    "TestUpdatePasswordSqlSmtp",
    "TestUpdateEmail",
    "TestUpdateEmailMongo",
    "TestGetUser",
    "TestGetUserMongo",
    "TestPasswordResets",
    "TestPasswordResetsMongo",
    "TestUsers",
    "TestUsersMongo",
    "TestDeleteAccount",
    "TestDeleteAccountMongo"
]



from .base import TestBase
from typing import cast
from .. import UnableToConfirmEmail, User


class TestEmailConfirm(TestBase):
    use_sql = True
    use_smtp = True

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_invalid_code(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            email="jamisbae@pm.me"
        )

        with self.assertRaises(UnableToConfirmEmail):
            await self.handler.confirm_email(
                email=cast(str, model.email),
                given_code="competelymadeup"
            )

    async def test_invalid_email(self) -> None:
        with self.assertRaises(UnableToConfirmEmail):
            await self.handler.confirm_email(
                email="googleisnsasimp@gmail.com",
                given_code="competelymadeup"
            )

    async def test_valid_confirm(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            email="lokiismorebae@gmail.com"
        )

        result = await self.handler._db_wrapper.get(
            "user", {"user_id": model.user_id}
        )
        if not result:
            raise Exception("this shouldn't happen")

        user = await self.handler.confirm_email(
            email=cast(str, model.email),
            given_code=result["email_vaildate"]
        )

        self.assertIsInstance(user, User)

        model = await user.get()
        self.assertTrue(model.email_confirmed)


class TestEmailConfirmMongo(TestEmailConfirm):
    use_sql = False
    use_smtp = True



from .base import TestBase
from .. import (
    PasswordPolicyError,
    AccountDetailsError,
    EmailError,
    NameLengthInvalidError,
    NameInvalidCharactersError,
    _MAX_NAME_LEN,
    UserModel,
    User,
    DetailsExistError
)


class TestCreateAccount(TestBase):
    use_sql = False
    use_smtp = False

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_password_policy_invalid(self) -> None:
        with self.assertRaises(PasswordPolicyError):
            await self.handler.create_account(
                password="1234hh$",
                name="ward"
            )

    async def test_no_email_or_name(self) -> None:
        with self.assertRaises(AccountDetailsError):
            await self.handler.create_account(
                self.valid_password
            )

    async def test_invalid_email(self) -> None:
        with self.assertRaises(EmailError):
            await self.handler.create_account(
                password=self.valid_password,
                email="something@.com"
            )

    async def test_invalid_names(self) -> None:
        with self.assertRaises(NameLengthInvalidError):
            await self.handler.create_account(
                password=self.valid_password,
                name="wa"
            )

        with self.assertRaises(NameLengthInvalidError):
            await self.handler.create_account(
                password=self.valid_password,
                name="x" * (_MAX_NAME_LEN + 1)
            )

        with self.assertRaises(NameInvalidCharactersError):
            await self.handler.create_account(
                password=self.valid_password,
                name="war$"
            )

    async def test_valid_account_only_email(self) -> None:
        model, user = await self.handler.create_account(
            password=self.valid_password,
            email="example@example.com"
        )

        self.assertIsInstance(model, UserModel)
        self.assertIsInstance(user, User)

    async def test_valid_account_only_name(self) -> None:
        model, user = await self.handler.create_account(
            password=self.valid_password,
            name="ward"
        )

        self.assertIsInstance(model, UserModel)
        self.assertIsInstance(user, User)

    async def test_valid_account_both(self) -> None:
        model, user = await self.handler.create_account(
            password=self.valid_password,
            name="notward",
            email="wardpearce@pm.me"
        )

        self.assertIsInstance(model, UserModel)
        self.assertIsInstance(user, User)

    async def test_account_name_taken(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            name="oopwe"
        )

        with self.assertRaises(DetailsExistError):
            await self.handler.create_account(
                password=self.valid_password,
                name=model.name
            )

    async def test_account_email_taken(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            email="oopwe@example.com"
        )

        with self.assertRaises(DetailsExistError):
            await self.handler.create_account(
                password=self.valid_password,
                email=model.email
            )


class TestCreateAccountSqlSmtp(TestCreateAccount):
    use_sql = True
    use_smtp = True



from typing import AsyncGenerator, Mapping, Optional
from sqlalchemy import (
    MetaData,
    Table,
    Column,
    Binary,
    String,
    Boolean,
    func,
    select,
    and_ as sql_and,
    or_ as sql_or,
    create_engine,
    TIMESTAMP
)

from databases import Database

from ._const import _MAX_NAME_LEN


metadata = MetaData()

user_table = Table(
    "user",
    metadata,
    Column(
        "user_id",
        String(length=32),
        primary_key=True
    ),
    Column(
        "name",
        String(length=_MAX_NAME_LEN),
        nullable=True
    ),
    Column(
        "email",
        String(length=255),
        nullable=True
    ),
    Column(
        "email_vaildate",
        String(length=43),  # secrets.token_urlsafe(32)
        nullable=True
    ),
    Column(
        "password_reset_code",
        String(length=43),  # secrets.token_urlsafe(32)
        nullable=True
    ),
    Column(
        "password_reset_generated",
        TIMESTAMP,
        nullable=True
    ),
    Column(
        "email_confirmed",
        Boolean(),
        nullable=True
    ),
    Column(
        "password",
        Binary()
    ),
    mysql_engine="InnoDB",
    mysql_charset="utf8mb4"
)


def create_tables(url: str) -> None:
    if "mysql" in url:
        old_engine = "mysql"
        engine = "pymysql"
    elif "sqlite" in url:
        old_engine = "sqlite"
        engine = "sqlite3"
    elif "postgresql" in url:
        old_engine = "postgresql"
        engine = "psycopg2"
    else:
        assert False, "Invalid database URL engine."

    metadata.create_all(
        create_engine(url.replace(old_engine, f"{old_engine}+{engine}"))
    )


class SqlWrapper:
    """Doesn't cover SQL syntax entirely, but just
       includes needed functions.
    """

    def __init__(self, db: Database) -> None:
        self._db = db

        self._tables = {
            "user": user_table
        }

    def __convert_to_clauses(self, table: Table, dict_: dict) -> list:
        return [
            table.c[key] == value for key, value in dict_.items()
        ]

    async def exists(self, table: str,
                     or_: dict) -> bool:
        return await self._db.fetch_val(
            select([func.count()]).select_from(
                self._tables[table]
            ).where(
                sql_or(*self.__convert_to_clauses(self._tables[table], or_))
            )
        ) != 0

    async def delete(self, table: str,
                     and_: dict) -> None:
        await self._db.execute(
            self._tables[table].delete().where(
                sql_and(*self.__convert_to_clauses(self._tables[table], and_))
            )
        )

    async def update(self, table: str,
                     and_: dict, values: dict) -> None:
        await self._db.execute(
            self._tables[table].update().values(
                **values
            ).where(
                sql_and(*self.__convert_to_clauses(self._tables[table], and_))
            )
        )

    async def insert(self, table: str,
                     values: dict) -> None:
        await self._db.execute(
            self._tables[table].insert().values(
                **values
            )
        )

    async def get(self, table, and_: dict) -> Optional[Mapping]:
        return await self._db.fetch_one(
            self._tables[table].select().where(
                sql_and(*self.__convert_to_clauses(self._tables[table], and_))
            )
        )

    async def iterate(self, table: str, and_: dict = None
                      ) -> AsyncGenerator[Mapping, None]:
        query = self._tables[table].select()

        if and_:
            query = query.where(
                sql_and(*self.__convert_to_clauses(self._tables[table], and_))
            )

        async for row in self._db.iterate(query):
            yield row



from .base import TestBase
from .. import AccountDetailsError, InvalidLogin


class TestLogin(TestBase):
    use_sql = False
    use_smtp = False

    valid_password = "S]Q}67=uLetG{r,_8{"

    async def test_no_name_or_email(self) -> None:
        with self.assertRaises(AccountDetailsError):
            await self.handler.login(password="12341")

    async def test_invalid_email(self) -> None:
        with self.assertRaises(InvalidLogin):
            await self.handler.login(password="12341", email="doens'texist")

    async def test_invalid_name(self) -> None:
        with self.assertRaises(InvalidLogin):
            await self.handler.login(
                password="12341",
                name="notarealname1231234"
            )

    async def test_valid_password_invalid_name(self) -> None:
        await self.handler.create_account(
            password=self.valid_password,
            name="somethingrllyc00l"
        )

        with self.assertRaises(InvalidLogin):
            await self.handler.login(
                password=self.valid_password,
                name="somethingrllyc"
            )

    async def test_valid_password_invalid_email(self) -> None:
        await self.handler.create_account(
            password=self.valid_password,
            email="somethingrllyc00l@pm.me"
        )

        with self.assertRaises(InvalidLogin):
            await self.handler.login(
                password=self.valid_password,
                email="somethingrllyc@pm.me"
            )

    async def test_valid_email_not_confirmed_with_smtp(self) -> None:
        if self.use_smtp:
            model, _ = await self.handler.create_account(
                password=self.valid_password,
                email="newemail@gamil.com"
            )

            with self.assertRaises(InvalidLogin):
                await self.handler.login(
                    email=model.email,
                    password=self.valid_password
                )

    async def test_valid_email_not_confirmed_without_smtp(self) -> None:
        if not self.use_smtp:
            model, _ = await self.handler.create_account(
                password=self.valid_password,
                email="newemail59@gamil.com"
            )

            await self.handler.login(
                email=model.email,
                password=self.valid_password
            )

    async def test_valid_name_invalid_password(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            name="wogamer420"
        )

        with self.assertRaises(InvalidLogin):
            await self.handler.login(
                password="1234",
                name=model.name
            )

    async def test_valid_email_invalid_password(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            email="somecreate@example.com"
        )

        with self.assertRaises(InvalidLogin):
            await self.handler.login(
                password="1234",
                email=model.email
            )

    async def test_valid_with_name(self) -> None:
        model, _ = await self.handler.create_account(
            password=self.valid_password,
            name="wogamer"
        )

        await self.handler.login(
            password=self.valid_password,
            name=model.name
        )

    async def test_valid_email_with_smtp(self) -> None:
        if self.use_smtp:
            model, _ = await self.handler.create_account(
                password=self.valid_password,
                email="wpearce6@gmail.com"
            )

            await self.handler._db_wrapper.update(
                "user", {"user_id": model.user_id},
                {"email_confirmed": True}
            )

            await self.handler.login(
                email=model.email,
                password=self.valid_password
            )


class TestLoginSqlSmtp(TestLogin):
    use_sql = True
    use_smtp = True



from typing import Mapping, Optional, TYPE_CHECKING
from bcrypt import checkpw, hashpw, gensalt
from email_validator import validate_email, EmailNotValidError
from secrets import token_urlsafe
from datetime import datetime

from ._models import UserModel
from ._errors import (
    UserIdError,
    InvalidLogin,
    EmailError,
    PasswordResetInvalid,
    DetailsExistError
)


if TYPE_CHECKING:
    from . import AccountHandler


class User:
    def __init__(self, upper: "AccountHandler", user_id: str) -> None:
        self._upper = upper
        self.user_id = user_id

    @property
    def __and(self) -> dict:
        return {"user_id": self.user_id}

    async def __raw_user(self) -> Optional[Mapping]:
        return await self._upper._db_wrapper.get(
            "user", self.__and
        )

    async def update_password(self, current_password: str,
                              new_password: str) -> None:
        """Used to update the password if they already know
           there current password.

        Parameters
        ----------
        current_password : str
        new_password : str

        Raises
        ------
        InvalidLogin
        PasswordPolicyError
        """

        user = await self.__raw_user()
        if not user:
            raise InvalidLogin()

        if not checkpw(current_password.encode(), user["password"]):
            raise InvalidLogin()

        self._upper._validate_details(password=new_password)

        await self._upper._db_wrapper.update(
            "user", self.__and, {
                "password": hashpw(new_password.encode(), gensalt())
            }
        )

    async def update_name(self, name: str) -> None:
        """Updates users name

        Parameters
        ----------
        name : str

        Raises
        ------
        DetailsExistError
        NameLengthInvalidError
        NameInvalidCharactersError
        """

        self._upper._validate_details(name=name)

        if await self._upper._db_wrapper.exists(
                "user", {"name": name}):
            raise DetailsExistError()

        await self._upper._db_wrapper.update(
            "user", self.__and, {"name": name}
        )

    async def update_email(self, new_email: str) -> None:
        """Used to update a user's email.

        Parameters
        ----------
        new_email : str

        Raises
        ------
        EmailError
        DetailsExistError
        """

        try:
            valid = validate_email(new_email)
        except EmailNotValidError as error:
            raise EmailError(str(error))

        if await self._upper._db_wrapper.exists(
                "user", {"email": valid.email}):
            raise DetailsExistError()

        values = {
            "email": valid.email,
        }

        if self._upper._smtp:
            values = {
                **values,
                **self._upper._email_regenerate()
            }

            await self._upper._jobs.spawn(
                self._upper._smtp._send(
                    values["email"],
                    values["email_vaildate"],
                    "confirm"
                )
            )

        await self._upper._db_wrapper.update(
            "user", self.__and, values
        )

    async def reset_password(self) -> str:
        """Used to reset a password

        Returns
        -------
        str
            The password reset code, if SMTP
            is being used you shouldn't need to touch this.
        """

        user = await self.get()

        values = {
            "password_reset_code": token_urlsafe(32),
            "password_reset_generated": datetime.now()
        }

        await self._upper._db_wrapper.update(
            "user", self.__and, values
        )

        if self._upper._smtp and user.email:
            await self._upper._jobs.spawn(
                self._upper._smtp._send(
                    user.email,
                    values["password_reset_code"],
                    "reset"
                )
            )

        return values["password_reset_code"]

    async def password_confirm(self, new_password: str,
                               given_code: str) -> None:
        """Used to reset password from password reset code.

        Parameters
        ----------
        new_password : str
        given_code : str

        Raises
        ------
        PasswordResetInvalid
        PasswordPolicyError
        """

        result = await self.__raw_user()
        if (not result or "password_reset_code" not in result or
                not result["password_reset_code"]):
            raise PasswordResetInvalid()

        if (datetime.now() - self._upper._password_reset_expires >
                result["password_reset_generated"]):
            raise PasswordResetInvalid()

        # Midigates timing attacks.
        valid = checkpw(
            given_code.encode(), hashpw(
                result["password_reset_code"].encode(),
                gensalt()
            )
        )

        if not valid:
            raise PasswordResetInvalid()

        self._upper._validate_details(password=new_password)

        await self._upper._db_wrapper.update(
            "user", self.__and, {
                "password": hashpw(new_password.encode(), gensalt()),
                "password_reset_code": None,
                "password_reset_generated": None
            }
        )

    async def get(self) -> UserModel:
        """Used to get details on user.

        Returns
        -------
        UserModel

        Raises
        ------
        UserIdError
        """

        result = await self.__raw_user()
        if not result:
            raise UserIdError()

        return UserModel(**result)

    async def delete(self) -> None:
        """Used to delete account, this can't be undone.
        """

        await self._upper._db_wrapper.delete("user", self.__and)



import os
import re

from setuptools import setup


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_long_description():
    with open("README.md", encoding="utf8") as f:
        return f.read()


def get_variable(variable):
    with open(os.path.join("aioaccount", "__init__.py")) as f:
        return re.search(
            "{} = ['\"]([^'\"]+)['\"]".format(variable), f.read()
        ).group(1)  # type: ignore


setup(
    name="aioaccount",
    version=get_variable("__version__"),
    url=get_variable("__url__"),
    description=get_variable("__description__"),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author=get_variable("__author__"),
    author_email=get_variable("__author_email__"),
    install_requires=get_requirements(),
    license=get_variable("__license__"),
    packages=[
        "aioaccount",
        "aioaccount.tests"
    ],
    python_requires=">=3.6",
    include_package_data=True,
    zip_safe=False
)



from .base import TestBase
from .. import UserModel, User


class TestUsers(TestBase):
    use_sql = True

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_users_listing(self) -> None:
        await self.handler.create_account(
            password=self.valid_password,
            name="jjd"
        )

        async for model, user in self.handler.users():
            self.assertIsInstance(model, UserModel)
            self.assertIsInstance(user, User)


class TestUsersMongo(TestUsers):
    use_sql = False



from password_strength import PasswordPolicy as ExtPP


class PasswordPolicy:
    def __init__(self, length: int = 8,
                 uppercase: int = 2,
                 numbers: int = 2,
                 special: int = 2,
                 nonletters: int = 2,
                 **kwargs) -> None:
        """Used to set password policy.

        Parameters
        ----------
        length : int, optional
            Min char length, by default 8
        uppercase : int, optional
            Min amount of uppercase chars required,
            by default 2
        numbers : int, optional
            Min amount of numbers required, by default 2
        special : int, optional
            Min amount of special chars required,
            by default 2
        nonletters : int, optional
            Min number of non letter characters,
            by default 2
        """

        self._policy = ExtPP.from_names(
            length=length,
            uppercase=uppercase,
            numbers=numbers,
            special=special,
            nonletters=nonletters,
            **kwargs
        )



from __future__ import annotations

import aiosmtplib

from email.mime.text import MIMEText
from email.message import EmailMessage
from jinja2 import Environment, FileSystemLoader, select_autoescape


class SmtpHtml:
    def __init__(self, path: str, file: str,
                 url_key: str = "url") -> None:
        """Configure SMTP html template

        Parameters
        ----------
        path : str
            Path to jinja2 templates.
        file : str
            Name of jinja2 template.
        url_key : str, optional
            Key for validation email url,
            by default "url"
        """

        self._jinja2 = Environment(
            loader=FileSystemLoader(path),
            autoescape=select_autoescape(["html", "xml"])
        )

        self._file = file
        self._url_key = url_key


class SmtpClient:
    def __init__(self, host: str, port: int, email: str,
                 **kwargs) -> None:
        """Used to configure SMTP.

        Parameters
        ----------
        host : str
        port : int
        email : str
            Email address to send as.

        Notes
        -----
        Kwargs are passed to aiosmtplib's
        SMTP client.

        https://aiosmtplib.readthedocs.io/en/stable/client.html
        """

        self._details = {
            "hostname": host,
            "port": port,
            **kwargs
        }

        self._email = email

        self._email_types = {
            "confirm": {
                "raw": "Please confirm your email\n{link}",
                "raw_place": True,
                "html": None,
                "subject": "Please confirm your email!",
                "url": "",
                "contains_code": False
            },
            "reset": {
                "raw": "Follow this link to reset your password\n{link}",
                "raw_place": True,
                "html": None,
                "subject": "Password reset request",
                "url": "",
                "contains_code": False
            }
        }

    def confirm_layout(self, url: str, html: SmtpHtml = None,
                       raw: str = None, subject: str = None
                       ) -> SmtpClient:
        """Used for confirm email layout.

        Parameters
        ----------
        url : str
            Url user follows for validation
            If it doesn't contain '{validation_code}'
            the validation code will be append
            to the end of the url

            e.g.
            https://example.com/validate?code={validation_code}
        html: SmtpHtml, optional
            by default None
        raw : str, optional
            Should contain 'link' otherwise
            appended at the end of the string,
            by default None
        subject : str, optional
            Used to set email subject for confirmations.
            by default None

        Returns
        -------
        SmtpClient
        """

        self._email_types["confirm"]["url"] = url
        self._email_types["confirm"][
            "contains_code"
        ] = "{validation_code}" in url

        if raw:
            self._email_types["confirm"]["raw"] = raw
            self._email_types["confirm"]["raw_place"] = (
                "{link}" in raw
                if raw else False
            )

        if html:
            self._email_types["confirm"]["html"] = html

        if subject:
            self._email_types["confirm"]["subject"] = subject

        return self

    def reset_layout(self, url: str, html: SmtpHtml = None,
                     raw: str = None, subject: str = None
                     ) -> SmtpClient:
        """Used for reset email layout.

        Parameters
        ----------
        url : str
            Url user follows for validation
            If it doesn't contain '{validation_code}'
            the validation code will be append
            to the end of the url

            e.g.
            https://example.com/validate?code={validation_code}
        html: SmtpHtml, optional
            by default None
        raw : str, optional
            Should contain 'link' otherwise
            appended at the end of the string,
            by default None
        subject : str, optional
            Used to set email subject for resets.
            by default None

        Returns
        -------
        SmtpClient
        """

        self._email_types["confirm"]["url"] = url
        self._email_types["confirm"][
            "contains_code"
        ] = "{validation_code}" in url

        if raw:
            self._email_types["reset"]["raw"] = raw
            self._email_types["reset"]["raw_place"] = (
                "{link}" in raw
                if raw else False
            )

        if html:
            self._email_types["reset"]["html"] = html

        if subject:
            self._email_types["reset"]["subject"] = subject

        return self

    async def _send(self, email: str, code: str, type_: str) -> None:
        """Used to send a email.

        Parameters
        ----------
        email : str
        code : str
        type_ : str
            email type
        """

        email_type = self._email_types[type_]

        if email_type["contains_code"]:
            link = email_type["url"].format(validation_code=code)
        else:
            link = email_type["url"] + code

        if email_type["html"]:
            message = MIMEText(email_type["html"]._jinja2.get_template(
                email_type["html"]._file
            ).render({email_type["html"]._url_key: link}), "html", "utf-8")
        else:
            message = EmailMessage()

            if email_type["raw"]:
                if email_type["raw_place"]:
                    content = email_type["raw"].format(link=link)
                else:
                    content = email_type["raw"] + link
            else:
                content = link

            message.set_content(content)

        message["From"] = self._email
        message["To"] = email
        message["Subject"] = email_type["subject"]

        await aiosmtplib.send(message, **self._details)



from .base import TestBase
from .. import InvalidLogin, PasswordPolicyError


class TestUpdatePassword(TestBase):
    use_sql = False
    use_smtp = False

    old_password = "#!K2&33?e%@Pv3_Q"
    new_password = "ZRvTP72g-VwX3@\\H"

    async def test_invalid_user(self) -> None:
        with self.assertRaises(InvalidLogin):
            await (self.handler.user("123")).update_password(
                current_password=self.old_password,
                new_password=self.new_password
            )

    async def test_valid_user_invalid_old_pass(self) -> None:
        _, user = await self.handler.create_account(
            password=self.old_password,
            name="deezemail"
        )

        with self.assertRaises(InvalidLogin):
            await user.update_password(
                current_password="1234",
                new_password=self.new_password
            )

    async def test_valid_user_invalid_new_pass(self) -> None:
        _, user = await self.handler.create_account(
            password=self.old_password,
            name="eeezz"
        )

        with self.assertRaises(PasswordPolicyError):
            await user.update_password(
                current_password=self.old_password,
                new_password="123"
            )

    async def test_valid_user(self) -> None:
        model, user = await self.handler.create_account(
            password=self.old_password,
            name="ppez"
        )

        await user.update_password(
            current_password=self.old_password,
            new_password=self.new_password
        )

        await self.handler.login(
            password=self.new_password,
            name=model.name
        )


class TestUpdatePasswordSqlSmtp(TestUpdatePassword):
    use_sql = True
    use_smtp = True



from .base import TestBase
from .. import AccountDetailsError, User, UserModel


class TestToUser(TestBase):
    use_sql = False
    use_smtp = False

    valid_password = "#!K2&33?e%@Pv3_Q"

    async def test_no_email_or_name(self) -> None:
        with self.assertRaises(AccountDetailsError):
            await self.handler.to_user()

    async def test_no_result(self) -> None:
        with self.assertRaises(AccountDetailsError):
            await self.handler.to_user(
                email="thisisn'treal"
            )

    async def test_valid_to_user_from_name(self) -> None:
        c_model, _ = await self.handler.create_account(
            password=self.valid_password,
            name="epicgamer1"
        )

        model, user = await self.handler.to_user(
            name=c_model.name
        )

        self.assertIsInstance(
            model,
            UserModel
        )

        self.assertIsInstance(
            user,
            User
        )

    async def test_valid_to_user_from_email(self) -> None:
        c_model, _ = await self.handler.create_account(
            password=self.valid_password,
            email="epicgamer@gmail.com"
        )

        model, user = await self.handler.to_user(
            email=c_model.email
        )

        self.assertIsInstance(
            model,
            UserModel
        )

        self.assertIsInstance(
            user,
            User
        )


class TestToUserSqlSmtp(TestToUser):
    use_sql = True
    use_smtp = True



from dataclasses import dataclass, fields
from typing import Union


@dataclass
class UserModel:
    """Holds details on a user.

    Attributes
    ----------
    user_id : str
    name : Union[None, str]
    email : Union[None, str]
    email_confirmed : Union[None, bool]
    """

    user_id: str
    name: Union[None, str] = None
    email: Union[None, str] = None
    email_confirmed: Union[None, bool] = None

    def __init__(self, **kwargs):
        names = set([f.name for f in fields(self)])
        for k, v in kwargs.items():
            if k in names:
                setattr(self, k, v)



class AioAccountError(Exception):
    """Base AioAccount Error
    """

    def __init__(self, msg: str, *args: object) -> None:
        super().__init__(msg, *args)


class AccountDetailsError(AioAccountError):
    pass


class EmailError(AccountDetailsError):
    pass


class NameLengthInvalidError(AccountDetailsError):
    pass


class NameInvalidCharactersError(AccountDetailsError):
    pass


class UnableToConfirmEmail(AccountDetailsError):
    def __init__(self, *args: object) -> None:
        super().__init__("Unable to confirm email with given details.", *args)


class PasswordResetInvalid(AccountDetailsError):
    def __init__(self, *args: object) -> None:
        super().__init__("Password reset code is invalid.", *args)


class DetailsExistError(AccountDetailsError):
    def __init__(self, *args: object) -> None:
        super().__init__("Account details already used.", *args)


class PasswordPolicyError(AccountDetailsError):
    def __init__(self, fails: list, *args: object) -> None:
        self.fails = fails
        super().__init__("Password doesn't meet password policy.", *args)


class InvalidLogin(AioAccountError):
    def __init__(self, *args: object) -> None:
        super().__init__("Provided details are incorrect.", *args)


class UserIdError(AioAccountError):
    def __init__(self, msg: str = "User id incorrect.", *args: object) -> None:
        super().__init__(msg, *args)
